/*
 * =========================================================================
 * Copyright (C) 1997 - 1998 by Visual Numerics, Inc. All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software is freely
 * granted by Visual Numerics, Inc., provided that the copyright notice
 * above and the following warranty disclaimer are preserved in human
 * readable form.
 *
 * Because this software is licenses free of charge, it is provided
 * "AS IS", with NO WARRANTY.  TO THE EXTENT PERMITTED BY LAW, VNI
 * DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO ITS PERFORMANCE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * VNI WILL NOT BE LIABLE FOR ANY DAMAGES WHATSOEVER ARISING OUT OF THE USE
 * OF OR INABILITY TO USE THIS SOFTWARE, INCLUDING BUT NOT LIMITED TO DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, PUNITIVE, AND EXEMPLARY DAMAGES, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. 
 *
 * =========================================================================
 */

package com.imsl.test.math;
import com.imsl.math.Complex;
import java.util.*;

abstract public class Test
{
	static final public double POSINF  = Double.POSITIVE_INFINITY;
	static final public double NEGINF  = Double.NEGATIVE_INFINITY;
	static final public double NaN     = Double.NaN;
	static final public double PI      = Math.PI;
	static final public double NEGZERO = 1/NEGINF;
	static final public int    REPEAT  = 20;

	final static private double TOLERANCE = 1.0e-13;
	static public boolean logPassingTests = true;
	static public boolean logFailingTests = true;
	static private java.io.PrintStream out = System.out;
	static private int nTestsPassed = 0;
	static private int nTestsFailed = 0;

	static private java.util.Random random = new java.util.Random(754321);

	abstract protected String getTestName();

	private void failHeader(String test)
	{
		out.println("Fail " + getTestName() + "." + test);
	}


	/**
	 *	Check the result of a test returning a Complex.
	 *	Test passes if either the absolute or releative error
	 *	is less than TOLERANCE.
	 *	@param	test	Name of the test for the log file.
	 *	@param	compute The computed value.
	 *	@param	expect	The expected value.
	 */
	protected boolean check(String test, Complex compute, Complex expect)
	{
		return check(test, compute, expect, TOLERANCE);
	}


	private boolean isNaN(Complex z)
	{
		return Double.isNaN(z.real()) || Double.isNaN(z.imag());
	}


	private boolean isInfinite(Complex z)
	{
		return Double.isInfinite(z.real()) || Double.isInfinite(z.imag());
	}


	/**
	 *	Check the result of a test returning a Complex.
	 *	Test passes if either the absolute or releative error
	 *	is less than tolerance.
	 *	@param	test	Name of the test for the log file.
	 *	@param	compute The computed value.
	 *	@param	expect	The expected value.
	 *	@param	tolerance	The tolerance.
	 *	@return	Returns true if test passes.
	 */
	protected boolean check(String test, Complex compute, Complex expect, double tolerance)
	{
		boolean pass;
		double	diff = 0.0;
		if (isInfinite(compute) || isInfinite(expect)
					|| isNaN(compute) || isNaN(expect)
					|| isNegZero(compute.real()) || isNegZero(compute.imag())
					|| isNegZero(expect.real())  || isNegZero(expect.imag())) {
			pass = (equals(compute.real(),expect.real()) && equals(compute.imag(),expect.imag()));
		} else {
			diff = Complex.abs(Complex.minus(compute,expect));
			double size = Complex.abs(expect);
			if (size < 1.0)  size = 1.0;
			diff /= size;
			pass = (diff < tolerance);
		}
		if (pass) {
			if (logPassingTests) {
				out.println("Pass "+test);
			}
			nTestsPassed++;
		} else {
			if (logFailingTests) {
				failHeader(test);
				out.println("   Computed "+compute.toString());
				out.println("   Expected "+expect.toString());
				if (diff > 0)
					out.println("   Relative error "+diff);
			}
			nTestsFailed++;
		}
		return pass;
	}


	private boolean equals(double x, double y)
	{
		if (Double.isNaN(x) &&  Double.isNaN(y))  return true;
		return (Double.doubleToLongBits(x)==Double.doubleToLongBits(y));
	}

	final private static long negZeroBits = Double.doubleToLongBits(NEGZERO);

	static final private boolean isNegZero(double x)
	{
		return (Double.doubleToLongBits(x) == negZeroBits);
	}


	protected boolean check(String test, double compute, double expect)
	{
		return check(test, compute, expect, TOLERANCE);
	}


	/**
	 *	Check the result of a test returning a double.
	 *	Test passes if either the absolute or releative error
	 *	is less than TOLERANCE.
	 *	@param	test	Name of the test for the log file.
	 *	@param	compute The computed value.
	 *	@param	expect	The expected value.
	 *	@param	tolerance	The tolerance. If tolerance is >= 1 then
	 *						it is in ulps.
	 */
	protected boolean check(String test, double compute, double expect, double tolerance)
	{
		boolean pass;
		double	diff = 0;
		if (Double.isInfinite(compute) || Double.isInfinite(expect)
			|| Double.isNaN(compute)   || Double.isNaN(expect)
			|| isNegZero(compute)	   || isNegZero(expect)) {
			pass = equals(compute,expect);
		} else if (tolerance >= 1) {
			long Lcompute = Double.doubleToLongBits(compute);
			long Lexpect  = Double.doubleToLongBits(expect);
			diff = Lcompute - Lexpect;
			//System.out.println(Long.toHexString(Lcompute)+"   "+Long.toHexString(Lexpect)+"  "+compute+"  "+diff);
			pass = (diff <= tolerance);
		} else {
			diff = Math.abs(compute-expect);
			double size = Math.abs(expect);
			if (size < 1.0)  size = 1.0;
			diff /= size;
			pass = (diff <= tolerance);
		}
		if (pass) {
			if (logPassingTests)  out.println("Pass "+test);
			nTestsPassed++;
		} else {
			if (logFailingTests) {
				Double Dcompute = new Double(compute);
				Double Dexpect  = new Double(expect);
				failHeader(test);
				long Lcompute = Double.doubleToLongBits(compute);
				long Lexpect  = Double.doubleToLongBits(expect);
				out.println("   Computed "+Long.toHexString(Lcompute)+"   "+Dcompute.toString());
				out.println("   Expected "+Long.toHexString(Lexpect)+"   "+Dexpect.toString());
				if (diff >= 1) {
					out.println("   Error "+diff+" ulp");
				} else if (diff >= 0) {
					out.println("   Relative error "+diff+"   "+(Lcompute-Lexpect));
				}
			}
			nTestsFailed++;
		}
		return pass;
	}

	protected boolean check(String test, double compute[], double expect[])
	{
		return check(test, compute, expect, TOLERANCE);
	}

	/**
	 *	Check the result of a test returning a double vector.
	 *	Test passes if either the absolute or releative error
	 *	is less than TOLERANCE.
	 *	@param	test	Name of the test for the log file.
	 *	@param	compute The computed vector.
	 *	@param	expect	The expected vector.
	 */
	protected boolean check(String test, double compute[], double expect[], double tolerance)
	{
		if (expect.length != compute.length) {
			if (logFailingTests) {
				failHeader(test);
				out.println("   Computed array nrows "+compute.length);
				out.println("   Expected array nrows "+expect.length);
			}
			nTestsFailed++;
			return false;
		}
	
		for (int i = 0;  i < expect.length;  i++) {
			boolean s = check(
				test+" element["+i+"]",
				compute[i], expect[i],
				tolerance);
			if (!s)  return s;
		}
		return true;
	}
	
	protected boolean check(String test, double compute[][], double expect[][])
	{
		return check(test, compute, expect, TOLERANCE);
	}

	/**
	 *	Check the result of a test returning a double array.
	 *	Test passes if either the absolute or releative error
	 *	is less than TOLERANCE.
	 *	@param	test	Name of the test for the log file.
	 *	@param	compute The computed array.
	 *	@param	expect	The expected array.
	 */
	protected boolean check(String test, double compute[][], double expect[][], double tolerance)
	{
		if (expect.length != compute.length) {
			if (logFailingTests) {
				failHeader(test);
				out.println("   Computed array nrows "+compute.length);
				out.println("   Expected array nrows "+expect.length);
			}
			nTestsFailed++;
			return false;
		}
	
		for (int i = 0;  i < expect.length;  i++) {
			for (int j = 0;  j < expect[i].length;  j++) {
				if (expect[i].length != compute[i].length) {
					if (logFailingTests) {
						failHeader(test);
						out.println("   Computed array for row "+i+" ncols="+compute[i].length);
						out.println("   Expected array for row "+i+" ncols="+expect[i].length);
					}
					nTestsFailed++;
					return false;
				}
				boolean s = check(
					test+" element["+i+"]["+j+"]",
					compute[i][j],
					expect[i][j],
					tolerance);
				if (!s)  return s;
			}
		}
		return true;
	}

	/**
	 *	Check the result of a test returning a long.
	 *	Test passes if the longs are identical
	 *	@param	test	Name of the test for the log file.
	 *	@param	compute The computed long.
	 *	@param	expect	The expected long.
	 */
	protected boolean check(String test, long compute, long expect)
	{
		boolean pass = (compute == expect);
		if (pass) {
			if (logPassingTests)  out.println("Pass "+test);
			nTestsPassed++;
		} else {
			if (logFailingTests) {
				failHeader(test);
				out.println("   Computed "+compute+" (0x"+Long.toHexString(compute)+"L)");
				out.println("   Expected "+expect+" (0x"+Long.toHexString(expect)+"L)");
			}
			nTestsFailed++;
		}
		return pass;
	}
	
	/**
	 *	Check the result of a test returning a String.
	 *	Test passes if the strings are identical
	 *	@param	test	Name of the test for the log file.
	 *	@param	compute The computed String.
	 *	@param	expect	The expected String.
	 */
	protected boolean check(String test, String compute, String expect)
	{
		if (compute.equals(expect)) {
			return true;
		} else {
			if (logFailingTests) {
				failHeader(test);
				out.println("   Computed "+compute);
				out.println("   Expected "+expect);
			}
			nTestsFailed++;
			return false;
		}
	}


	/**
	 *	Check the result of a test returning an Exception.
	 *	Test passes if the strings are identical
	 *	@param	test	Name of the test for the log file.
	 *	@param	compute The computed String.
	 *	@param	expect	The expected String.
	 */
	protected boolean check(String test, Exception compute, Exception expect)
	{
		//if (compute.equals(expect)) {
		if (compute.getMessage().equals(expect.getMessage())) {
			return true;
		} else {
			if (logFailingTests) {
				failHeader(test);
				out.println("   Computed "+compute);
				out.println("   Expected "+expect);
			}
			nTestsFailed++;
			return false;
		}
	}

	/**
	 *	Return a random number.
	 *	We do not use Math.random() so that we can explictly set the
	 *	seed so that all runs generate the same results.
	 */
	static protected double random()
	{
		return random.nextDouble();
	}

	/**
	 *	Set the seed for Test.random()
	 *	Test classes that use random() should explicitly set the seed
	 *	to insulate themselves from the effects of other tests.
	 */
	static protected void setSeed(long seed)
	{
		random.setSeed(seed);
	}

	/**
	 *	Wait for input.
	 *	Used if results window would be destroyed before being read.
	 */
	static private void pause()
	{
		try {
			System.in.read();
		} catch (Exception e) {
		}
	}

	public void all()
	{
	}

	public void test()
	{
		int failed = nTestsFailed;
		all();
		failed = nTestsFailed - failed;
		System.out.println("Failed " + failed +
			((failed < 10) ? "  " : " ") +
			getTestName());
	}


	public static void main(String args[])
	{
		System.out.println("Vendor:  "+System.getProperties().getProperty("java.vendor"));
		System.out.println("Version: "+System.getProperties().getProperty("java.version"));
		System.out.println("OS Name: "+System.getProperties().getProperty("os.name"));
		System.out.println("OS Arch: "+System.getProperties().getProperty("os.arch"));
		System.out.println("OS Vers: "+System.getProperties().getProperty("os.version"));
		
		logPassingTests = false;
		logFailingTests = true;

		try {
			new TestMath().test();
			//new TestJMath().test();
			new TestComplex().test();
			new TestSfun().test();
		} catch (Exception exception) {
			exception.printStackTrace();
		}
		
		System.out.println("Test passing " + nTestsPassed);
		System.out.println("Test failing " + nTestsFailed);
		//pause();
	}
}
